#ifndef MOVIE_HPP
#define MOVIE_HPP

#include "config.h"

#include "module.hpp"
#include "options.hpp"
#include "config.hpp"
#include "gettext.hpp"

#include "movie_config.hpp"

#include "busy_indicator.hpp"

#include "multifile.hpp"
#include "common-feature.hpp"
#include "libfspp.hpp"
#include "plugins.hpp"

#include "boost.hpp"

#ifdef use_inotify
#include "notify.hpp"
#endif

#include <stack>
#include <vector>
#include <list>

class Movie : public Module
{
protected:
  Movie();

  bool search_compare(const Multifile& s);
  std::string get_name_from_file(const Multifile& e);

  // commands
  virtual void options() = 0;
  void exit();
  virtual void search_func() = 0;

  // helper function
  void add_standard(ExtraMenu& em);

  bool exit_loop;

  enum SUBTYPE { IDX, SRT, NOSUB };

  // returns a pair of type, path_to_subtitles
  std::pair<SUBTYPE, std::vector<std::string> > find_subtitles(const std::string& dir);

  bool printing_information;

  // playback
  void playmovie(const Multifile file, bool window = false);
  void external_movie_playback(const std::string& path);
  void play_dvd_helper(const std::string& device);
  void playcd(int count);
  void playvcd();
  void playdvd();

  // list of folders from config, the names are normalized to have / at the end
  std::list<std::string> movie_folders;

#ifdef use_inotify
  virtual void fs_change(NotifyUpdate::notify_update_type type, const std::string& path) = 0;
#endif
  virtual bool reload_dir(const std::string& path) = 0;
  virtual void reload_current_dirs() = 0;

  std::string testdir(int *count);

  Multifile addfile(const std::string& name, const MyPair& filetype);
  Multifile add_dir(const std::string& filename);

  virtual void load_runtime_settings();

  // For traversing the filesystem. The first element is the dirs and the next is the
  // position in the list generated from the dirs
  std::stack<std::pair<std::list<std::string>, int > > folders;

  class PlayerPluginOpts: public Options
  {
  public:
    void add_opt(Option *_opt);
    void set_header(const std::string &_str){header =_str;};
    void save(){    }; //do not save from here
  };

  class MovieOpts : public Options
  {
  public:
    MovieOpts();
    ~MovieOpts();

    std::string dir_order();

    void dummy(const std::string& str);
    std::string movie_player();
    std::string vcd_player();
    std::string dvd_player();

    std::string display_mode();

    bool mainloop();

    void set_players_correctly();
    bool vcd_constructed;
    bool dvd_constructed;

    // evil hack
    void clear();

  private:
    bool players_set_correctly;

    Option *reload_p;
    Option *dir_order_p;
    Option *movie_player_p, *vcd_player_p, *dvd_player_p;
    Option *display_mode_p;
  };

  struct file_sort
  {
    bool operator()(const Multifile& lhs, const Multifile& rhs);
  };

  virtual void load_current_dirs() = 0;

  // control player
  void add_playback_keys();
  bool active_control_player();
  void player_play();
  void player_pause();
  void player_stop();
  void player_ff();
  void player_fb();
  void player_fullscreen();

  bool playback_command(const Input& i);
  bool fullscreen_check(const Input& i);

  MoviePlayer *active_player;

  bool movie_playing;
  bool movie_playback_shown;

  MovieConfig *movie_conf;

public:

  void set_folders();

  virtual void save_runtime_settings();

#ifdef use_inotify
  // notify
  virtual void reparse_current_dir() = 0;
#endif

  MovieOpts* get_opts() { return &opts; }

  MovieOpts opts;

  void set_movie_playback(bool playback)
  {
    movie_playing = playback;
  }

  virtual void read_dirs() = 0;
  virtual void determine_media() = 0;
  virtual int files_size() = 0;

  void startup_updater();

  void reset();
};

template <typename T>
class MovieTemplate : public Movie
{
protected:

  void print_lcd_menu()
  {
    if (!global->lcd_output_possible())
      return;

    // dir
    std::string dir = "/";
    if (folders.size() != 1) {
      std::stack<std::pair<std::list<std::string>, int > > tmp_folders = folders;

      std::string name = tmp_folders.top().first.front();

      for (int number_of_slashes = tmp_folders.size(); number_of_slashes > 0; --number_of_slashes) {
	int p = name.rfind('/');
	if (p != std::string::npos) {
	  if (number_of_slashes != tmp_folders.size())
	    dir = name.substr(p) + dir;
	  name = name.substr(0, p);
	} else
	  break;
      }
    }

    // current
    std::string cur_name = "> " + get_name_from_file(vector_lookup(files, folders.top().second));

    std::string prev_name = "";

    if (files.size() > 2) {
      // prev
      int prev_pos = folders.top().second - 1;
      if (prev_pos == -1)
	prev_pos = files.size()-1;

      prev_name = get_name_from_file(vector_lookup(files, prev_pos));
    }

    std::string next_name = "";

    if (files.size() > 1) {
      // next
      int next_pos = folders.top().second + 1;
      if (next_pos == files.size())
	next_pos = 0;

      next_name = get_name_from_file(vector_lookup(files, next_pos));
    }

    // FIXME: lcd size
    //    LcdConf *lcd_conf = LcdConf::get_instance();

    // FIXME: display the last part of the string instead of this stuff

    global->lcd_add_output(string_format::pretty_print_string_copy(dir, 20));
    global->lcd_add_output(string_format::pretty_print_string_copy(prev_name, 20));
    global->lcd_add_output(string_format::pretty_print_string_copy(cur_name, 20));
    global->lcd_add_output(string_format::pretty_print_string_copy(next_name, 20));
    global->lcd_print();
  }

  void playcd(int count)
  {
    if (count == 1) { //play disc

      std::list<MyPair> filetypes = movie_conf->p_filetypes_m();
      std::string file_types;

      foreach (MyPair& filetype, filetypes)
	file_types += "-iname '*." + filetype.first + "' -o ";

      file_types = file_types.substr(0, file_types.size()-4);

      std::string play_cmd = "find " + cd->get_mount_point() +
	" -follow -maxdepth 8 -type f -printf \"'%p' \" " +
	file_types + " | xargs ";

      Plugins *plugins = S_Plugins::get_instance();
      MoviePlayerPlugin *plugin = plugins->find(plugins->movie_players, opts.movie_player());

      if (plugin->player->wants_exclusivity())
	render->device->unlock();

      plugin->player->play_disc(play_cmd);

      if (plugin->player->wants_exclusivity())
	render->device->lock();

    } else {

      std::vector<T> backup_files;

      backup_files = files;

      std::stack<std::pair<std::list<std::string>, int > > backup_folders = folders;

      std::list<std::string> dirs;
      dirs.push_back(cd->get_mount_point());

      files = parse_dir(dirs);

      while (!folders.empty())
	folders.pop();

      folders.push(std::make_pair(dirs, 0));

      mainloop();

      files = backup_files;
      folders = backup_folders;
    }

    //umount the disc
    run::external_program("umount " + cd->get_mount_point());
  }

  virtual std::vector<T> parse_dir(const std::list<std::string>& dirs) = 0;

  // commands
  void action_play()
  {
    Multifile e = vector_lookup(files, folders.top().second);
    int nfiles = e.filenames.size();
    if (nfiles > 0)
      playmovie(e);
    else
      DialogWaitPrint pdialog(dgettext("mms-movie", "Folder is empty"), 1000);
  }

  virtual void page_down() = 0;
  virtual void page_up() = 0;

#ifdef use_inotify
  void update_db_current_dirs()
  {
    foreach (std::string& dir, folders.top().first)
      reload_dir(dir);
  }
#endif

  void options()
  {
    visible = false;

    S_BusyIndicator::get_instance()->idle();

    if (opts.mainloop()) {

#ifdef use_inotify
      update_db_current_dirs();

      reparse_current_dir();
#else
      reload_current_dirs();

      // sanity check
      if (folders.top().second > files.size()-1)
	folders.top().second = 0;
#endif
    }

    opts.save();

    visible = true;
  }

  void movie_mainloop_common(const Input& input)
  {
    if (input.command == "page_up")
      {
	page_up();
      }
    else if (input.command == "page_down")
      {
	page_down();
      }
    else if (input.command == "options")
      {
	options();
      }
    else if (input.command == "startmenu")
      {
	exit();
      }
    else if (input.command == "switch_fullscreen")
      {
	if (render->device->supports_wid() && movie_conf->p_playback_in_mms())
	  player_fullscreen();
      }
  }

  std::vector<T> files;

public:
  int files_size() { return files.size(); }

  void load_current_dirs()
  {
    files = parse_dir(folders.top().first);

    if (folders.top().first.size() > 1 && folders.size() == 1)
      std::sort(files.begin(), files.end(), file_sort());
  }

  virtual void check_for_changes() {}

  void read_dirs()
  {
    // process movie dir
    files = parse_dir(movie_folders);

    if (movie_folders.size() > 1)
      std::sort(files.begin(), files.end());

    S_BackgroundUpdater::get_instance()->run_once(boost::bind(&MovieTemplate<T>::check_for_changes, this));
  }

  virtual void print(const std::vector<T>& files_to_print) = 0;

#ifdef use_inotify
  // notify
  void fs_change(NotifyUpdate::notify_update_type type, const std::string& path)
  {
    if (type == NotifyUpdate::MODIFY_FILE)
      ; // FIXME: doesn't work because inotify returns dir and not the file: reparse_file(path);
    else
      reload_dir(path);

    foreach (std::string& dir, folders.top().first)
      if (path == dir) {
	reparse_current_dir();
	break;
      }

    if (type == NotifyUpdate::DELETE_DIR) {
      bool current_dir_ok = false;

      while (!current_dir_ok) {
	load_current_dirs();
	if (files.size() == 0) {
	  if (folders.size() == 1) {
	    exit();
	    break;
	  }
	  else
	    folders.pop();
	} else
	  current_dir_ok = true;
      }
    }

    // sanity check
    if (folders.top().second > files.size()-1)
      folders.top().second = files.size()-1;

    if (!(active_control_player() && active_player->fullscreen()) &&
	!global->playback_in_fullscreen && !printing_information && visible)
      print(files);
  }

  void reparse_current_dir()
  {
    reload_current_dirs();

    // sanity check
    if (folders.top().second > files.size()-1)
      folders.top().second = 0;
  }
#endif

  void determine_media()
  {
    bool data_disc = cd->data_disc();
    cd->close();

    if (data_disc) {
      //mount the disc
      run::external_program("mount " + cd->get_mount_point());

      int count;
      std::string mediatype = testdir(&count);

      if (mediatype == "empty")
	Print pdialog(dgettext("mms-movie", "The disc does not contain any recognizable files"),
		      Print::SCREEN);
      else if (count < 2)
	DialogWaitPrint pdialog(dgettext("mms-movie", "Starting playback..."), 1000);

      if (mediatype != "divx")
	//umount the disc
	run::external_program("umount " + cd->get_mount_point());

      if (mediatype == "vcd")
	playvcd();
      else if (mediatype == "svcd")
	playvcd();
      else if (mediatype == "dvd")
	playdvd();
      else if (mediatype == "divx")
	playcd(count);
    } else
      Print pdialog(dgettext("mms-movie", "The disc does not contain any recognizable files"),
		    Print::SCREEN);
  }
};

#endif
