// for prefix
#include "config.h"

#include "input.hpp"

#include "config.hpp"
#include "common.hpp"

#include "global.hpp"
#include "updater.hpp"

#include "replay.hpp"

#include "busy_indicator.hpp"

#include "touch.hpp"

#include "gettext.hpp"

#include <fstream>
#include <sstream>
#include <iostream>

using std::string;

pthread_mutex_t InputMaster::singleton_mutex = PTHREAD_MUTEX_INITIALIZER;

string Input::to_string() const
{
  return mode + "," + key + "," + command + "," + name + "," + conv::btos(multiple);
}

Input::Input(const string& c, const string m)
  : multiple(false)
{
  command = c;
  mode = m;
}

Input::Input()
  : multiple(false)
{}

Input::Input(string& s)
{
  string::size_type comma = s.find(",");
  mode = s.substr(0, comma);
  s = s.substr(comma+1);

  comma = s.find(",");
  key = s.substr(0, comma);
  s = s.substr(comma+1);

  comma = s.find(",");
  command = s.substr(0, comma);
  s = s.substr(comma+1);

  comma = s.find(",");
  name = s.substr(0, comma);
  s = s.substr(comma+1);

  multiple = conv::stob(s);
}

void InputDevice::register_keypress(Input input)
{
  // volume changes should not count as a keypress
  if ((input.mode != "playback" || input.command == "playback_commands") &&
      (input.mode != "audio_general" ||
       (input.command == "playlist_mode" || input.command == "mode"))) {
    // update time of last keypress
    S_Config::get_instance()->s_last_key(time(0));
  }
}

void InputDevice::register_cancel_callback(const boost::function<void (void)>& callback)
{
  cancel_callback = callback;
}

void InputDevice::clear_cancel_callback()
{
  cancel_callback = 0;
}

void InputDevice::run_cancel_callback()
{
  if (cancel_callback != 0)
    cancel_callback();
}

void InputDevice::trigger_input(Input input)
{
  S_ScreenUpdater::get_instance()->trigger.trigger_input(input.command);
}

Input InputDevice::get_input()
{
  S_Global::get_instance()->reset_stop_bit();

  input_done_.wait();
  input_done_.reset();
  input_ready.signal();

  register_keypress(input);

  // FIXME: check if plugin is available?
  ReplaySaver *r = S_ReplaySaver::get_instance();
  r->add_event(input);

  return input;
}

void InputDevice::busy_idle(BusyIndicator *indicator)
{
  indicator->idle();
}

Input InputDevice::get_input_busy_wrapped(BusyIndicator *indicator)
{
  indicator->idle();
  Input return_value = get_input();
  indicator->busy();
  return return_value;
}

InputDevice::InputDevice()
  : accept_all(false), touch_action(false), cancel_callback(0)
{}

InputMaster::InputMaster()
{
  parse_keys("default", "4");
  parse_keys("search", "3");
}

string InputMaster::find_shortcut(const string& command)
{
  foreach (const Input& input, cur_list())
    if (command == input.command)
      return input.presentation;

  std::cerr << "something wrong:" << command << " not found" << std::endl;

  return "";
}

bool InputMaster::check_version(const string& input, const string& str,
				const string& version)
{
  // first line must be a version line
  string::size_type p = str.find("=");
  string param = str.substr(0, p-1);
  string value = str.substr(p+2);

  if (param == "version") {
    if (value == version) {
      return true;
    } else {
      DebugPrint perror(Print::DEBUGGING, DebugPrint::CRITICAL, "INPUT");
      perror.add_line(gettext("Wrong version of input file found: ") + input);
      perror.add_line(gettext("Please upgrade your input file"));
      perror.add_line(gettext("Got version ") + value + gettext(" expected ") + version);
      perror.print();
      exit(0);
    }
  } else {
    DebugPrint perror(Print::DEBUGGING, DebugPrint::CRITICAL, "INPUT");
    perror.add_line(gettext("Version string not found"));
    perror.add_line(gettext("Please upgrade your input file"));
    perror.print();
    exit(0);
  }

  return false;
}

std::pair<bool, std::pair<input_map, input_list> > InputMaster::parse_config(const string& plugin,
									     const string& version,
									     bool extends)
{
  Config *config = S_Config::get_instance();

  string input;
  string _str= (extends) ? "_extend" : "";
  input = config->p_homedir() + "/input/" + config->p_input() + "/" + plugin + _str;

  if (!file_exists(input))
    input =  std::string(mms_config) + "/input/" + config->p_input() + "/" + plugin + _str;

  input_map imap;
  input_list ilist;

  std::ifstream In(input.c_str());   //create an input file stream

  if(!In){
    print_critical(gettext("Input configuration file not found ") + input, "INPUT");
    return std::make_pair(false, std::make_pair(imap, ilist));
  } else {

    string str;

    bool version_correct = false;

    while (getline(In, str)) {
      // skips empty and strings which starts with a #
      if (str[0] == '#' || str.size() == 0)
	continue;

      if (!version_correct) {
	version_correct = check_version(input, str, version);
	if (!version_correct)
	  return std::make_pair(false, std::make_pair(imap, ilist));
      }

      string::size_type pos = str.find(",");

      if (pos == string::npos) // bogus line
	continue;

      string mode = str.substr(0, pos);

      string strlast = str.substr(pos+1);
      string::size_type pos2 = strlast.find(",");

      if (pos2 == string::npos) // bogus line
	continue;

      string strlast2 = strlast.substr(pos2+1);
      string::size_type pos3 = strlast2.find(",");

      string command = str.substr(pos+1, pos2);

      string key, mult;

      if (pos3 != string::npos) {
	key = strlast.substr(pos2+1, pos3); // pos3 is relative to strlast2
	mult = strlast2.substr(pos3+1);
      } else {
	key = strlast.substr(pos2+1);
	mult = "";
      }

      string presentation;
      string::size_type pos4 = key.find("|");
      if (pos4 == string::npos) { // no presentation defined
        presentation = key;
      } else {
        presentation = key.substr(pos4+1);
        key = key.substr(0,pos4);
      }

      bool multiple = false;

      if (mult == "yes")
	multiple = true;

      Input putback;

      putback.mode = mode;
      putback.command = command;
      putback.key = key;
      putback.presentation = presentation;
      putback.multiple = multiple;

      print_debug(gettext("Adding ") + mode + " " + command + " " + key + ":" + presentation + " " + putback.name, "INPUT");

      imap.insert(std::make_pair(putback.key, putback));
      ilist.push_back(putback);
    }

    return std::make_pair(true, std::make_pair(imap, ilist));
  }
}

//const input_desc_map& desc,
bool InputMaster::parse_keys(const string& plugin, const string& version)
{
  std::pair<bool, std::pair<input_map, input_list> > result = parse_config(plugin, version, false);

  // FIXME: could it just not return an empty list?

  if (!result.first)
    return false;

  input_map imap = result.second.first;
  input_list ilist = result.second.second;

  if (plugin == "default")
    maps.insert(std::make_pair(plugin, std::make_pair(imap, ilist)));
  else {
    std::map<string, std::pair<input_map, input_list> >::iterator element = maps.find("default");
    if (element == maps.end())
      std::cerr << "warning input loaded for plugin before loading main input" << std::endl;
    else {
      // add default values
      std::pair<input_map, input_list> input = element->second;
      imap.insert(input.first.begin(), input.first.end());
      ilist.insert(ilist.end(), input.second.begin(), input.second.end());
    }
    maps.insert(std::make_pair(plugin, std::make_pair(imap, ilist)));
  }

  return true;
}

bool InputMaster::restore_map_empty()
{
  return saved_map.empty();
}

void InputMaster::save_map()
{
  saved_map = cur_map;
}

void InputMaster::restore_map()
{
  cur_map = saved_map;
}

void InputMaster::set_map(const string& name)
{
  cur_map = name;
}

string InputMaster::current_saved_map()
{
  return saved_map;
}

string InputMaster::current_map()
{
  return cur_map;
}

bool InputMaster::find_input(const string& input, Input &i)
{
  name_input_map::iterator iter = maps.find(cur_map);

  if (iter != maps.end()) {
    input_map cmap = iter->second.first;
    input_map::iterator l = cmap.find(input);
    if (l != cmap.end()) {
      i = l->second;
      return true;
    } else
      return false;
  } else {
    print_warning(gettext("Unknown cur_map: ") + cur_map + "!", "INPUT");
    return false;
  }
}

bool InputMaster::extend_default(const string& plugin, const string& version)
{
  std::pair<bool, std::pair<input_map, input_list> > result = parse_config(plugin, version, true);

  if (!result.first)
    return false;

  input_map imap = result.second.first;
  input_list ilist = result.second.second;

  name_input_map new_maps;

  foreach (const name_input& element, maps) {
    if (element.first != "search") {
      input_map imap_temp = element.second.first;
      imap_temp.insert(imap.begin(), imap.end());
      input_list ilist_temp =element.second.second;
      ilist_temp.insert(ilist_temp.end(), ilist.begin(), ilist.end());
      new_maps.insert(std::make_pair(element.first, std::make_pair(imap_temp, ilist_temp)));
    } else {
      new_maps.insert(element);
    }
  }

  maps = new_maps;

  return true;
}

input_list InputMaster::cur_list()
{
  name_input_map::iterator i = maps.find(cur_map);
  if (i != maps.end())
    return i->second.second;
  else {
    print_warning(gettext("Unknown cur_map: ") + cur_map + "!", "INPUT");
    input_list l;
    return l;
  }
}
