#include "config.h"

#include "sdlout.hpp"

#include "touch.hpp"

#ifdef benchmark_output
#include "bench.hpp"
#endif

#include "print.hpp"
#include "gettext.hpp"

#include <sstream>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/kd.h>
#include <sys/types.h>
#include <stdlib.h>

#ifdef use_nls
#include <langinfo.h>
#include <locale.h>
#endif

void SDLdev::deinit()
{
  if (initialized == -1)
    return;

  SDL_ShowCursor(SDL_ENABLE);

  initialized = -1;

  if(conf->p_fullscreen())
    toggle_fullscreen();

  locked_mutex.leaveMutex();
}

void SDLdev::init()
{
  if(SDL_Init(SDL_INIT_VIDEO) < 0) {
    print_critical(dgettext("mms-sdl", "Could not initialize") + std::string(" ") + SDL_GetError(), "SDL");
    std::exit(1);
  }

  screen = SDL_SetVideoMode(width, height, 32, SDL_SWSURFACE);

  if(!screen) {
    print_critical(dgettext("mms-sdl", "Could not set video mode ") +
		      conv::itos(width) + "x" + conv::itos(height) +
		      "x32," + dgettext("mms-sdl", "error") + " " + SDL_GetError(), "SDL");
    std::exit(1);
  }

  rgbsurface = SDL_CreateRGBSurface(SDL_SRCCOLORKEY, width, height, 32,
				    0x00FF0000, 0x0000FF00, 0x000000FF, 0/*0xFF000000*/);


  // set the caption
  SDL_WM_SetCaption("mms", "mms");

  bpp = screen->format->BitsPerPixel;
  line_len = screen->format->BytesPerPixel * width;

#ifdef use_x11
  SDL_SysWMinfo info;
  SDL_VERSION(&info.version);

  int parent;
  if (SDL_GetWMInfo(&info) > 0) {
    info.info.x11.lock_func();
    display = info.info.x11.display;
    sdl_window_id = info.info.x11.wmwindow;
    parent = info.info.x11.window;
    int x11_width = DisplayWidth(info.info.x11.display, 0);
    int x11_height = DisplayHeight(info.info.x11.display, 0);
    info.info.x11.unlock_func();
    if (x11_width < width || x11_height < height) {
      print_critical("MMS resolution larger than X11 resolution", "SDL");
      ::exit(1);
    }
  }

  window_id = XCreateSimpleWindow(display, parent, width-width/6, height-height/6, width/6, height/6, 0, 0, 0);
  XUnmapWindow(display, window_id);

  fs_window_id = XCreateSimpleWindow(display, parent, 0, 0, width, height, 0, 0, 0);
  XUnmapWindow(display, fs_window_id);
#endif

  if(conf->p_fullscreen())
    toggle_fullscreen();

  render = S_Render::get_instance();

  initialized = 1;
}

bool SDLdev::supports_wid()
{
#ifdef use_x11
  return true;
#else
  return false;
#endif
}

int SDLdev::wid()
{
  return window_id;
}

void SDLdev::show_window_video()
{
  if (!window_mapped) {
#ifdef use_x11
    XMapWindow(display, window_id);
#endif
    window_mapped = true;
  }
}

void SDLdev::hide_window_video()
{
  if (window_mapped) {
#ifdef use_x11
    XUnmapWindow(display, window_id);
#endif
    window_mapped = false;
  }
}

int SDLdev::fs_wid()
{
  return fs_window_id;
}

void SDLdev::show_fs_video()
{
  if (!fs_window_mapped) {
#ifdef use_x11
    XMapWindow(display, fs_window_id);
#endif
    fs_window_mapped = true;
  }
}

void SDLdev::hide_fs_video()
{
  if (fs_window_mapped) {
#ifdef use_x11
    XUnmapWindow(display, fs_window_id);
#endif
    fs_window_mapped = false;

    locked_mutex.enterMutex();
    SDL_PollEvent(reinterpret_cast<SDL_Event *>(0));
    locked_mutex.leaveMutex();

    render->complete_redraw();
  }
}

void SDLdev::hide_video()
{
  hide_window_video();
  hide_fs_video();
}

void SDLdev::lock()
{
  locked_mutex.leaveMutex();

  if(conf->p_fullscreen())
    toggle_fullscreen();

  render->complete_redraw();
}

void SDLdev::unlock()
{
  if(conf->p_fullscreen())
    toggle_fullscreen();

  locked_mutex.enterMutex();
}

SDLdev::SDLdev()
  : width(conf->p_h_res()), height(conf->p_v_res()),
    initialized(-1), verbose(3), fullscreen(false),
    window_mapped(false), fs_window_mapped(false)
{
#ifdef use_nls
  // gettext
  setlocale(LC_ALL, ""); /* set from the environment variables */
  bindtextdomain("mms-sdl", mms_prefix"/share/locale");
  bind_textdomain_codeset("mms-sdl", nl_langinfo(CODESET));
#endif
}

SDLdev::~SDLdev()
{
  terminate(); /* thread */
}

void SDLdev::draw_rectangle(int x, int y, int w, int h)
{
  SDL_Rect rect;
  rect.x = x;
  rect.y = y;

  if (w > conf->p_h_res())
    rect.w = conf->p_h_res();
  else
    rect.w = w;

  if (h > conf->p_v_res())
    rect.h = conf->p_v_res();
  else
    rect.h = h;

#if 0
  std::ostringstream out;
  out << "blitting " << rect.x << ":" << rect.y << ":" << rect.w << ":" << rect.h;
  DebugPrint print(out.str(), Print::INFO, DebugPrint::ALL, "SDL out");
#endif

  SDL_BlitSurface(rgbsurface, &rect, screen, &rect);
  SDL_UpdateRect(screen, rect.x, rect.y, rect.w, rect.h);
}

void SDLdev::sdldev_draw_32()
{
  rgbsurface->pixels = reinterpret_cast<UINT8 *>(imlib_image_get_data_for_reading_only());

  foreach (rect& r, render->draw_rects)
    draw_rectangle(r.x, r.y, r.w, r.h);
}

void SDLdev::sdldev_draw()
{
  if (render->draw_rects.size() == 0) // optimization
    return;

  locked_mutex.enterMutex();
  switch (bpp) {
  case 32 : sdldev_draw_32(); break;
    //  case 24 : sdldev_draw_24(dataptr); break;
    //  case 16 : sdldev_draw_16(dataptr); break;
  default :
    {
      print_critical("Color depth not supported", "SDL");
    }
  }
  locked_mutex.leaveMutex();
}

//----------------------------------------------------------------------
void SDLdev::run()
{
#ifdef benchmark_output
  Bench bench;
#endif

  while (true) {

    drawing = false;
    new_to_draw = false;

    output_done.signal();

    output_ready.wait();
    output_ready.reset();

    drawing = true;

#ifdef benchmark_output
    bench.start();
#endif

    sdldev_draw();

#ifdef benchmark_output
    bench.stop();
    bench.print("Rendering using SDL");
#endif
  }
}

void SDLdev::redraw()
{
  sdldev_draw();
}

void SDLdev::toggle_fullscreen()
{
  locked_mutex.enterMutex();

  SDL_LockSurface(screen);

  int opts = SDL_SWSURFACE;

  if (!fullscreen)
    opts |= SDL_FULLSCREEN;

  screen = SDL_SetVideoMode(width, height, 32,  opts);

  SDL_UnlockSurface(screen);

  locked_mutex.leaveMutex();

  if (fullscreen) {

    fullscreen = false;

    render->prepare_new_image();
    render->clear();
    render->draw("clear");
    render->image_mut.leaveMutex();

  } else
    fullscreen = true;

  locked_mutex.enterMutex();

  SDL_LockSurface(screen);

  SDL_ShowCursor((screen->flags & SDL_FULLSCREEN) > 0 ? SDL_DISABLE:SDL_ENABLE);

  SDL_UnlockSurface(screen);

  locked_mutex.leaveMutex();
}
