#include "config.h"

#include "plugins.hpp"

#include "libfspp.hpp"

#include "boost.hpp"

#include <signal.h>
#include <stdio.h>
#include <dlfcn.h>
#include <sys/stat.h>

#include <fstream>
#include <iostream>

using std::vector;
using std::string;

pthread_mutex_t Plugins::singleton_mutex = PTHREAD_MUTEX_INITIALIZER;

void Plugins::delete_plugins()
{
  foreach (FeaturePlugin* plugin, fp_data) {
    //    void *handle = plugin->handle;
    plugin->destructor(plugin);
    //    dlclose(handle);
  }
  foreach (AudioPlayerPlugin* plugin, audio_players) {
    //    void *handle = plugin->handle;
    plugin->destructor(plugin);
    //    dlclose(handle);
  }
  foreach (MoviePlayerPlugin* plugin, movie_players) {
    //    void *handle = plugin->handle;
    plugin->destructor(plugin);
    //    dlclose(handle);
  }
  foreach (PrinterPlugin* plugin, printers) {
    void *handle = plugin->handle;
    plugin->destructor(plugin);
    dlclose(handle);
  }
  foreach (InputPlugin* plugin, ip_data) {
    void *handle = plugin->handle;
    plugin->destructor(plugin);
    dlclose(handle);
  }
  foreach (OutputPlugin* plugin, op_data) {
    //    void *handle = plugin->handle;
    if (plugin->loaded()) {
      plugin->device->deinit();
    }
    plugin->destructor(plugin);
    // FIXME: probably something to do with sdl_quit
    //    dlclose(handle);
  }
}

Plugin *Plugins::add_plugin(const string& filename)
{
  print_debug("Plugins::add_plugin() " + filename);

  string base_filename = basename(filename.c_str());

  print_debug("Plugins name is " + base_filename);

  // search in all types if already loaded plugin

  // input plugin ?
  foreach (InputPlugin* plugin, ip_data)
    if (base_filename == basename(plugin->filename.c_str()))
      return NULL;

  print_debug("Not an already loaded input plugin");

  // output plugin ?
  foreach (OutputPlugin* plugin, op_data)
    if (base_filename == basename(plugin->filename.c_str()))
      return NULL;

  print_debug("Not an already loaded output plugin");

  // feature plugin ?
  foreach (FeaturePlugin* plugin, fp_data)
    if (base_filename == basename(plugin->filename.c_str()))
      return NULL;

  print_debug("Not an already loaded feature plugin");

  // audio player plugin ?
  foreach (AudioPlayerPlugin* plugin, audio_players)
    if (base_filename == basename(plugin->filename.c_str()))
      return NULL;

  print_debug("Not an already loaded audio player plugin");

  // movie player plugin ?
  foreach (MoviePlayerPlugin* plugin, movie_players)
    if (base_filename == basename(plugin->filename.c_str()))
      return NULL;

  print_debug("Not an already loaded movie player plugin");

  // printer plugin ?
  foreach (PrinterPlugin* plugin, printers)
    if (base_filename == basename(plugin->filename.c_str()))
      return NULL;

  print_debug("Not an already loaded print plugin");

  print_debug("Not already loaded, thus loading plugin");

  Plugin *p = load(filename);
  return p;
}

Plugin* Plugins::load(const string& filename)
{
  void *h;
  void *(*constructor) (void);
  void *(*destructor) (Plugin*);

  string base_filename = basename(filename.c_str());

  if ((h = dlopen(filename.c_str(), RTLD_NOW | RTLD_GLOBAL))) {

    print_debug("Plugin loaded");

    // FIXME oldstyle cast - ISO C++ forbid it directly - howto do this now ??
    if ((constructor = (void *(*)())dlsym(h, "construct")) && ((destructor = (void *(*)(Plugin*))dlsym(h, "destroy"))))
      {
	Plugin *p = static_cast<Plugin *>(constructor());
	if (p) {
	  p->handle = h;
	  p->destructor = destructor;
	  p->filename = filename;
	  switch (p->type()) {
	  case IN_PLUGIN:
	  case OUT_PLUGIN:
	  case FEATURE_PLUGIN:
	  case AUDIO_PLAYER_PLUGIN:
	  case MOVIE_PLAYER_PLUGIN:
	  case PRINTER_PLUGIN:
	    return p;
	    break;
	  default:
	    print_warning("Unknown plugin type, removed: " + base_filename);
	    delete p;
	    break;
	  }
	} else
	  print_warning("Plugin fail to pre-init: " + base_filename);
      } else
      print_warning("Does not look like a MMS plugin");
  } else
    print_warning(string("dlopen failed with error: ") + dlerror());

  dlclose(h);

  return NULL;
}

bool prioritize_plugins(const FeaturePlugin *fp1, const FeaturePlugin *fp2)
{
  return fp1->priority() < fp2->priority();
}

void Plugins::scan_input_output()
{
  string dir = MMS_PLUGINS_DIR;

  print_debug("Plugins scanning: " + dir);

  for (file_iterator<file_t, default_order> i (dir); i != i.end (); i.advance(false)) {

    string filename = i->getName();

    print_debug("Plugins Scan: trying " + filename);

    int p;
    struct stat statbuf;

    if (!stat(filename.c_str(), &statbuf) && S_ISREG(statbuf.st_mode))
      if ((p=filename.rfind('.'))>0 && (filename.substr(p) == ".so")) {

	// hack to not load plugins we don't neeed
	if ((filename.find("lib_ainput_") == string::npos) &&
	    (filename.find("lib_output_") == string::npos))
	  continue;

        Plugin *plugin = add_plugin(filename);
        if (plugin) {
	  switch (plugin->type()) {
	  case IN_PLUGIN:{
	    InputPlugin *ip = static_cast<InputPlugin *>(plugin);
	    print_info("Input plugin: " + ip->plugin_name() + " loaded");
	    ip_data.push_back(ip);
	    break;
	  }
	  case OUT_PLUGIN:{
	    OutputPlugin *op = static_cast<OutputPlugin *>(plugin);
	    print_info("Output plugin: " + op->plugin_name() + " loaded");
	    op_data.push_back(op);
	    break;
	  }
	  default:
	    break;
	  }
	} else
	  print_warning("Error loading plugin: " + filename);
      }
  }
}

void Plugins::scan()
{
  string dir = MMS_PLUGINS_DIR;

  print_debug("Plugins scanning: " + dir);

  for (file_iterator<file_t, default_order> i (dir); i != i.end (); i.advance(false)) {

    string filename = i->getName();

    print_debug("Plugins Scan: trying " + filename);

    int p;
    struct stat statbuf;

    if (!stat(filename.c_str(), &statbuf) && S_ISREG(statbuf.st_mode))
      if ((p=filename.rfind('.'))>0 && (filename.substr(p) == ".so")) {
        Plugin *plugin = add_plugin(filename);
        if (plugin) {
	  switch (plugin->type()) {
	  case IN_PLUGIN:{
	    InputPlugin *ip = static_cast<InputPlugin *>(plugin);
	    print_info("Input plugin: " + ip->plugin_name() + " loaded");
	    ip_data.push_back(ip);
	    break;
	  }
	  case OUT_PLUGIN:{
	    OutputPlugin *op = static_cast<OutputPlugin *>(plugin);
	    print_info("Output plugin: " + op->plugin_name() + " loaded");
	    op_data.push_back(op);
	    break;
	  }
	  case FEATURE_PLUGIN:{
	    FeaturePlugin *fp = static_cast<FeaturePlugin *>(plugin);
	    if (fp->items().size() > 0) {
	      print_info("Feature plugin: " + fp->plugin_name() + " loaded");
	      fp_data.push_back(fp);
	    }
	    break;
	  }
	  case AUDIO_PLAYER_PLUGIN:{
	    AudioPlayerPlugin *ap = static_cast<AudioPlayerPlugin *>(plugin);
	    print_info("Audio player plugin: " + ap->plugin_name() + " loaded");
	    audio_players.push_back(ap);
	    break;
	  }
	  case MOVIE_PLAYER_PLUGIN:{
	    MoviePlayerPlugin *vp = static_cast<MoviePlayerPlugin *>(plugin);
	    print_info("Movie player plugin: " + vp->plugin_name() + " loaded");
	    movie_players.push_back(vp);
	    break;
	  }
	  case PRINTER_PLUGIN:{
	    PrinterPlugin *pp = static_cast<PrinterPlugin *>(plugin);
	    print_info("Printer plugin: " + pp->plugin_name() + " loaded");
	    printers.push_back(pp);
	    break;
	  }
	  default:
	    print_warning("Unknown plugin type: " + filename);
	  }
	} else
	  print_warning("Error loading plugin: " + filename);
      }
  }

  // sort feature plugins by priority
  std::sort(fp_data.begin(), fp_data.end(), prioritize_plugins);
}

Plugin* Plugins::find(const string& name)
{
  foreach (FeaturePlugin* plugin, fp_data)
    if (name == plugin->plugin_name())
      return plugin;
  foreach (OutputPlugin* plugin, op_data)
    if (name == plugin->plugin_name())
      return plugin;
  foreach (InputPlugin* plugin, ip_data)
    if (name == plugin->plugin_name())
      return plugin;
  foreach (AudioPlayerPlugin* plugin, audio_players)
    if (name == plugin->plugin_name())
      return plugin;
  foreach (MoviePlayerPlugin* plugin, movie_players)
    if (name == plugin->plugin_name())
      return plugin;
  foreach (PrinterPlugin* plugin, printers)
    if (name == plugin->plugin_name())
      return plugin;
  return 0;
}
