#include "updater.hpp"

#include "config.hpp"

#include "boost.hpp"

// time
#include <time.h>

#include <string>

#include <iostream>

using std::string;
using std::list;

pthread_mutex_t ScreenUpdater::singleton_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t BackgroundUpdater::singleton_mutex = PTHREAD_MUTEX_INITIALIZER;

Updater::Updater()
  : progressbar(Overlay("progressbar")), status_progressbar(0),
    total_progressbar(0), header(""), running(false), quitting(false)
{}

void Updater::activate()
{
  running = true;
}

void Updater::deactivate()
{
  running = false;
}

void Updater::quit()
{
  timer.elements_mut.enterMutex();
  running = true;
  quitting = true;
  timer.elements_mut.leaveMutex();
}

void Updater::run_once(const boost::function<void (void)>& func)
{
    timer.elements_mut.enterMutex();
    run_once_functions.push_back(func);
    timer.elements_mut.leaveMutex();
}

void Updater::run()
{
  set_nicelevel();

  while (true) {
    while (!running)
      sleep(500); // ms

    timer.elements_mut.enterMutex();
    if (quitting){
      timer.elements_mut.leaveMutex();
      break;
    }

    foreach (boost::function<void (void)>& func, run_once_functions)
      func();

    run_once_functions.clear();

    gettimeofday(&timer.current_time, NULL);

    long int next_timeout = 1000;
    long int min_timeout = 250; // at least sleep 250 ms

    foreach (TimeElement& element, timer.elements) {
      if (element.active) {

	struct timeval diff = diff_timeval(element.next_time, timer.current_time);

	if (diff.tv_sec < 0 || diff.tv_usec < 0) {
	  element.callback();
	  element.next_time = timer.set_timeout(timer.current_time, element.time());
	}

	next_timeout = std::min(next_timeout,
				(element.next_time.tv_sec-timer.current_time.tv_sec)*1000+
				(element.next_time.tv_usec-timer.current_time.tv_usec)/1000);
      }
    }

    if (next_timeout < min_timeout)
      next_timeout = min_timeout;

    timer.elements_mut.leaveMutex();

    sleep(next_timeout); // ms
  }
}

struct timeval Updater::diff_timeval(const struct timeval& t1, const struct timeval& t2)
{
  struct timeval diff;

  if (t1.tv_usec > t2.tv_usec){
    diff.tv_usec = t1.tv_usec - t2.tv_usec;
    diff.tv_sec = t1.tv_sec - t2.tv_sec;
  } else {
    diff.tv_usec = t1.tv_usec + 1000000 - t2.tv_usec;
    diff.tv_sec = t1.tv_sec - t2.tv_sec - 1;
  }

  return diff;
}

struct timeval Updater::Timer::set_timeout(const struct timeval& t, int wait)
{
  struct timeval temp;
  temp.tv_sec = t.tv_sec + wait/1000;
  temp.tv_usec = t.tv_usec + (wait - ((wait/1000)*1000)) * 1000;
  if (temp.tv_usec > 1000*1000) {
    temp.tv_sec += 1;
    temp.tv_usec -= 1000*1000;
  }
  return temp;
}

void Updater::Timer::add(TimeElement te)
{
  elements_mut.enterMutex();

  gettimeofday(&current_time, NULL);
  te.next_time = set_timeout(current_time, te.time());
  elements.push_back(te);

  elements_mut.leaveMutex();
}

void Updater::Timer::del(string name)
{
  elements_mut.enterMutex();

  for (list<TimeElement>::iterator iter = elements.begin(), end = elements.end();
       iter != end; ++iter)
    if (iter->name == name) {
      elements.erase(iter);
      break;
    }

  elements_mut.leaveMutex();
}

void Updater::Timer::activate(const string& name)
{
  elements_mut.enterMutex();

  foreach (TimeElement& element, elements)
    if (element.name == name) {
      gettimeofday(&current_time, NULL);
      element.next_time.tv_sec = current_time.tv_sec + element.time()/1000;
      element.next_time.tv_usec = current_time.tv_usec + (element.time() - (element.next_time.tv_sec*1000)) * 1000;
      element.active = true;
      break;
    }

  elements_mut.leaveMutex();
}

void Updater::Timer::deactivate(const string& name)
{
  elements_mut.enterMutex();

  foreach (TimeElement& element, elements)
    if (element.name == name) {
      element.active = false;
      break;
    }

  elements_mut.leaveMutex();
}

void Updater::Timer::run_and_deactivate(const string& name)
{
  elements_mut.enterMutex();

  foreach (TimeElement& element, elements)
    if (element.name == name) {
      element.callback();
      element.active = false;
      break;
    }

  elements_mut.leaveMutex();
}

bool Updater::Timer::status(const string& name)
{
  elements_mut.enterMutex();

  bool active_element = false;
  foreach (TimeElement& element, elements)
    if (element.name == name) {
      active_element = element.active;
      break;
    }

  elements_mut.leaveMutex();

  return active_element;
}

// trigger

void Updater::Trigger::add(const TriggerElement& te)
{
  elements_mut.enterMutex();

  elements.push_back(te);

  elements_mut.leaveMutex();
}

void Updater::Trigger::del(const string& name)
{
  elements_mut.enterMutex();

  for (list<TriggerElement>::iterator iter = elements.begin(), end = elements.end();
       iter != end; ++iter)
    if (iter->name == name) {
      if (iter->cleanup)
	iter->cleanup();
      elements.erase(iter);
      break;
    }

  elements_mut.leaveMutex();
}

void Updater::Trigger::trigger_input(const string& trigger)
{
  if (!enabled)
    return;

  elements_mut.enterMutex();

  foreach (TriggerElement& element, elements) {
    if (element.triggers.size() == 1 && element.triggers.front().empty()) { // trigger
								      // on all input
      element.trigger();
    } else {
      foreach (string& the_trigger, element.triggers)
	if (the_trigger == trigger) {
	  element.trigger();
	  break;
	}
    }
  }

  elements_mut.leaveMutex();
}

void Updater::Trigger::enable()
{
  enabled = true;
}

void Updater::Trigger::disable()
{
  enabled = false;
}

Updater::Trigger::Trigger()
  : enabled(true)
{}

void BackgroundUpdater::set_nicelevel()
{
  Config *conf = S_Config::get_instance();

  if (conf->p_priority_change())
    nice(10); //Lower priority to keep audio from skipping
}

void ScreenUpdater::set_nicelevel()
{
  Config *conf = S_Config::get_instance();

  if (conf->p_priority_change())
    nice(3); //Lower priority to keep audio from skipping
}
